/*******************************************************************************
 * C H A M E L E O N   ColdFire C file                                         *
 *******************************************************************************
 * Copyright (C) 2001-2002 Soundart                                            *
 * www.soundart-hot.com                                                        *
 * support@soundart-hot.com                                                    *
 ******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <rtems.h>
#include <chameleon.h>

#include "dsp/dsp_code.h"

/******************************************************************************/
#define WORKSPACE_SIZE	128*1024
rtems_unsigned32 rtems_workspace_size = WORKSPACE_SIZE;
rtems_unsigned32 rtems_workspace_start[WORKSPACE_SIZE];
/******************************************************************************/

#define kSampleRate		48000
#define MAX_PARAMETERS_BY_PAGE	10
#define	MAX_PAGES_BY_GROUP	10
#define MAX_GROUPS_BY_MENU	10

enum
{
	kParam_Early1Delay = 0,
	kParam_Early1Gain,
	kParam_Early2Delay,
	kParam_Early2Gain,
	kParam_Early3Delay,
	kParam_Early3Gain,

	kParam_CombReverbTime,
	kParam_CombHiDamping,
	kParam_Comb1Delay,
	kParam_Comb2Delay,
	kParam_Comb3Delay,
	kParam_Comb4Delay,
	kParam_Comb5Delay,
	kParam_Comb6Delay,

	kParam_Allpass1Delay,
	kParam_Allpass2Delay,
	kParam_Allpass3Delay,

	kParam_Mix,
	kParam_Brightness,
	kParam_Volume,

	kParamCount
};

struct menu_page_t
{
	const char *		menu_name;
	int 			param_count;
	int			param_selected;
	int			params[MAX_PARAMETERS_BY_PAGE];
};

struct menu_group_t
{
	const char *		menu_name;
	int 			page_count;
	int			page_selected;
	struct menu_page_t	pages[MAX_PAGES_BY_GROUP];
};

struct menu_t
{
	int			group_count;
	int			group_selected;
	struct menu_group_t	groups[MAX_GROUPS_BY_MENU];
};

static struct menu_t menu =
{
	4,
	0,
	{{
		"Early Reflect ",
		3,
		0,
		{{
			"1",
			2,
			0,
			{	
				kParam_Early1Delay,
				kParam_Early1Gain,
			}
		},{
			"2",
			2,
			0,
			{
				kParam_Early2Delay,
				kParam_Early2Gain,
			}
		},{
			"3",
			2,
			0,
			{
				kParam_Early3Delay,
				kParam_Early3Gain,
			}
		}},
	},{
		"Comb Filter",
		7,
		0,
		{{
			" all",
			2,
			0,
			{
				kParam_CombReverbTime,
				kParam_CombHiDamping,
			}
		},{
			" 1  ",
			1,
			0,
			{
				kParam_Comb1Delay,
			}
		},{
			" 2  ",
			1,
			0,
			{
				kParam_Comb2Delay,
			}
		},{
			" 3  ",
			1,
			0,
			{
				kParam_Comb3Delay,
			}
		},{
			" 4  ",
			1,
			0,
			{
				kParam_Comb4Delay,
			}
		},{
			" 5  ",
			1,
			0,
			{
				kParam_Comb5Delay,
			}
		},{
			" 6  ",
			1,
			0,
			{
				kParam_Comb6Delay,
			}
		}}
	},{
		"Allpass Filter",
		3,
		0,
		{{
			"1",
			1,
			0,
			{
				kParam_Allpass1Delay,
			}
		},{
			"2",
			1,
			0,
			{
				kParam_Allpass2Delay,
			}
		},{
			"3",
			1,
			0,
			{
				kParam_Allpass3Delay,
			}
		}}
	},{
		"Global          ",
		1,
		0,
		{{
			"",
			2,
			0,
			{
				kParam_Mix,
				kParam_Brightness,
			}
		}}
	}}
};

float parameters[kParamCount] =
{
	20.0f, 		/* kParam_Early1Delay */
	0.78f,		/* kParam_Early1Gain */
	38.9f,		/* kParam_Early2Delay */
	0.635f,		/* kParam_Early2Gain */
	69.9f,		/* kParam_Early3Delay */
	0.267f,		/* kParam_Early3Gain */

	5.0f,		/* kParam_CombReverbTime */
	10.0f,		/* kParam_CombHiDamping */
	50.0f,		/* kParam_Comb1Delay */
	56.0f,		/* kParam_Comb2Delay */
	61.0f,		/* kParam_Comb3Delay */
	68.0f,		/* kParam_Comb4Delay */
	72.0f,		/* kParam_Comb5Delay */
	78.0f,		/* kParam_Comb6Delay */

	6.0f,		/* kParam_Allpass1Delay */
	1.7f,		/* kParam_Allpass2Delay */
	0.63f,		/* kParam_Allpass3Delay */

	0.4f,		/* kParam_Mix */
	0.5f,		/* kParam_Brightness */
	0.0f,		/* kParam_Volume */
};

void get_param_value(int param_id, char *param_value)
{
	switch (param_id)
	{
		case kParam_Early1Delay:
		case kParam_Early2Delay:
		case kParam_Early3Delay:
		case kParam_CombReverbTime:
		case kParam_CombHiDamping:
		case kParam_Comb1Delay:
		case kParam_Comb2Delay:
		case kParam_Comb3Delay:
		case kParam_Comb4Delay:
		case kParam_Comb5Delay:
		case kParam_Comb6Delay:
		case kParam_Allpass1Delay:
		case kParam_Allpass2Delay:
		case kParam_Allpass3Delay:
			sprintf(param_value, "%.1f", parameters[param_id]);
			break;

		case kParam_Early1Gain:
		case kParam_Early2Gain:
		case kParam_Early3Gain:
		case kParam_Brightness:
		case kParam_Mix:
			sprintf(param_value, "%.2f", parameters[param_id]);
			break;
		
		default:
			sprintf(param_value, " ");
	}
}

const char *get_param_name(int param_id)
{
	switch (param_id)
	{
		case kParam_Early1Delay:
		case kParam_Early2Delay:
		case kParam_Early3Delay:
		case kParam_Comb1Delay:
		case kParam_Comb2Delay:
		case kParam_Comb3Delay:
		case kParam_Comb4Delay:
		case kParam_Comb5Delay:
		case kParam_Comb6Delay:
		case kParam_Allpass1Delay:
		case kParam_Allpass2Delay:
		case kParam_Allpass3Delay:
			return "Delay(ms)";

		case kParam_Early1Gain:
		case kParam_Early2Gain:
		case kParam_Early3Gain:
			return "Gain     ";

		case kParam_CombReverbTime:
			return "RevTime  ";

		case kParam_CombHiDamping:
			return "Damping  ";

		case kParam_Brightness:
			return "Bright   ";

		case kParam_Mix:
			return "Mix      ";
	}
	
	return "";
}

void change_param_value(int param_id, int increment)
{
	float min, max, scale;
	
	switch (param_id)
	{
		case kParam_Early1Delay:
		case kParam_Early2Delay:
		case kParam_Early3Delay:
		case kParam_Comb1Delay:
		case kParam_Comb2Delay:
		case kParam_Comb3Delay:
		case kParam_Comb4Delay:
		case kParam_Comb5Delay:
		case kParam_Comb6Delay:
			min = 0.1f;
			max = 85.0f;
			scale = 0.1f;
			break;

		case kParam_Allpass1Delay:
		case kParam_Allpass2Delay:
		case kParam_Allpass3Delay:
			min = 0.1f;
			max = 10.0f;
			scale = 0.1f;
			break;

		case kParam_Early1Gain:
		case kParam_Early2Gain:
		case kParam_Early3Gain:
		case kParam_Brightness:
		case kParam_Mix:
			min = 0.0f;
			max = 1.0f;
			scale = 0.01f;
			break;

		case kParam_CombReverbTime:
			min = 0.1f;
			max = 20.0f;
			scale = 0.05f;
			break;

		case kParam_CombHiDamping:
			min = 1.0f;
			max = 50.0f;
			scale = 0.1f;
			break;
	}
	
	parameters[param_id] += (float) increment*scale;
	
	if (parameters[param_id] > max)
		parameters[param_id] = max;
	else if (parameters[param_id] < min)
		parameters[param_id] = min;
}

void error(char *error_msg)
{
	TRACE(error_msg);
	exit(1);
}

int isprime(int val)
{
	int i, max = (int) (sqrtf(val) + 0.5f);
	
	for (i=3; i<max ;i+=2)
	{
		if ((val % i) == 0)
			return 0;
	}
	
	return 1;
}

int sampdel(float delay)
{
	int lag = (int) (delay * 0.001f * kSampleRate + 0.5f);
	
	lag += (lag % 2) - 1;
	
	while (isprime(lag) == 0)
		lag = lag + 2;
	
	return lag;
}

void getcoec(float delay, float time, float damping, float *coec, float *coed)
{
	float dly, coel, coeh;
	
	dly = delay * 0.001f / time;
	coel = powf(0.001f, dly);
	coeh = powf(0.001f, dly * damping);
	*coed = (coel - coeh) / (coel + coeh);
   	*coec = (1.0f - (*coed)) * coel;
}

static rtems_task dsp_task(rtems_task_argument argument)
{
	rtems_signed32 data[31];
	rtems_event_set	pending;
	rtems_status_code status;
	float time, damping, coec, coed;
	int dsp;

	TRACE("Starting dsp_task...\n");

	dsp = dsp_init(1, dspCode);
	if (!dsp)
		error("ERROR: cannot access the DSP.\n");

	while (TRUE)
	{
		status = rtems_event_receive(
			RTEMS_EVENT_1,
			RTEMS_WAIT | RTEMS_EVENT_ANY,
			RTEMS_NO_TIMEOUT,
			&pending
		);
		if (status != RTEMS_SUCCESSFUL)
			break;

		TRACE("Updating DSP parameters...");
		fflush(stdout);

		data[ 0] = (rtems_signed32) sampdel(parameters[kParam_Early1Delay]);
		data[ 1] = float_to_fix(parameters[kParam_Early1Gain]);
		data[ 2] = (rtems_signed32) sampdel(parameters[kParam_Early2Delay]);
		data[ 3] = float_to_fix(parameters[kParam_Early2Gain]);
		data[ 4] = (rtems_signed32) sampdel(parameters[kParam_Early3Delay]);
		data[ 5] = float_to_fix(parameters[kParam_Early3Gain]);
	
		time = parameters[kParam_CombReverbTime];
		damping = parameters[kParam_CombHiDamping];
	
		getcoec(parameters[kParam_Comb1Delay], time, damping, &coec, &coed);
		data[ 6] = (rtems_signed32) sampdel(parameters[kParam_Comb1Delay]);
		data[ 7] = float_to_fix(coec);
		data[ 8] = float_to_fix(coed);
		getcoec(parameters[kParam_Comb2Delay], time, damping, &coec, &coed);
		data[ 9] = (rtems_signed32) sampdel(parameters[kParam_Comb2Delay]);
		data[10] = float_to_fix(coec);
		data[11] = float_to_fix(coed);
		getcoec(parameters[kParam_Comb3Delay], time, damping, &coec, &coed);
		data[12] = (rtems_signed32) sampdel(parameters[kParam_Comb3Delay]);
		data[13] = float_to_fix(coec);
		data[14] = float_to_fix(coed);
		getcoec(parameters[kParam_Comb4Delay], time, damping, &coec, &coed);
		data[15] = (rtems_signed32) sampdel(parameters[kParam_Comb4Delay]);
		data[16] = float_to_fix(coec);
		data[17] = float_to_fix(coed);
		getcoec(parameters[kParam_Comb5Delay], time, damping, &coec, &coed);
		data[18] = (rtems_signed32) sampdel(parameters[kParam_Comb5Delay]);
		data[19] = float_to_fix(coec);
		data[20] = float_to_fix(coed);
		getcoec(parameters[kParam_Comb6Delay], time, damping, &coec, &coed);
		data[21] = (rtems_signed32) sampdel(parameters[kParam_Comb6Delay]);
		data[22] = float_to_fix(coec);
		data[23] = float_to_fix(coed);
	
		data[24] = (rtems_signed32) sampdel(parameters[kParam_Allpass1Delay]);
		data[25] = (rtems_signed32) sampdel(parameters[kParam_Allpass2Delay]);
		data[26] = (rtems_signed32) sampdel(parameters[kParam_Allpass3Delay]);
	
		data[27] = float_to_fix(parameters[kParam_Brightness]);
		data[28] = float_to_fix(parameters[kParam_Mix]);
		data[29] = float_to_fix(parameters[kParam_Volume]);
		data[30] = 0xFFFFFF;
		
		if (!dsp_write_data(dsp, data, 31))
			break;
			
		TRACE("\n");
	}

	dsp_exit(dsp);

	TRACE("Ending dsp_task...\n");
	rtems_task_delete(RTEMS_SELF);
}

rtems_id create_dsp_task(void)
{
	rtems_id		task_id;
	rtems_status_code	status;

	status = rtems_task_create(
		rtems_build_name('D', 'S', 'P', ' '),
		10,
		RTEMS_MINIMUM_STACK_SIZE,
		RTEMS_DEFAULT_MODES,
		RTEMS_DEFAULT_ATTRIBUTES,
		&task_id
	);

	if (status != RTEMS_SUCCESSFUL)
	{
		TRACE("ERROR: cannot create dsp_task.\n");
		return 0;
	}

	status = rtems_task_start(task_id, dsp_task, 0);
	if (status != RTEMS_SUCCESSFUL)
	{
		TRACE("ERROR: cannot start dsp_task.\n");
		return 0;
	}

	return task_id;
}

static void display(int panel)
{
	struct menu_group_t *	group = &menu.groups[menu.group_selected];
	struct menu_page_t *	page = &group->pages[group->page_selected];
	int 			param_id = page->params[page->param_selected];
	static char 		line0[16+1];
	static char 		line1[16+1];

	char param_value[16];
	
	get_param_value(param_id, param_value);

	sprintf(line0, "%s:%s", group->menu_name, page->menu_name);
	sprintf(line1, "%s [%4s]", get_param_name(param_id), param_value);

	panel_out_lcd_print(panel, 0, 0, line0);
	panel_out_lcd_print(panel, 1, 0, line1);
}

rtems_timer_service_routine timer_routine(rtems_id timer_id, void *user_data)
{
	rtems_event_send((rtems_id) user_data, RTEMS_EVENT_1);
}

static rtems_task panel_task(rtems_task_argument argument)
{
	static float		volume_table[128];
	struct menu_group_t *	group = &menu.groups[menu.group_selected];
	struct menu_page_t *	page = &group->pages[group->page_selected];
	int 			param_id = page->params[page->param_selected];
	rtems_unsigned8		potentiometer;
	rtems_unsigned8		value;
	rtems_unsigned32	key_bits;
	rtems_unsigned8		encoder;
	rtems_signed8		increment;
	rtems_status_code	status;
	rtems_id		timer_id;
	int			i, panel;
	float			dB;

	TRACE("Starting panel_task...\n");

	/* Fill volume_table with a variable gain from -60 dB to 0 dB */
	for (i=0; i<128 ;i++)
	{
		if (i < 27)
			dB = -90.0 + (float) 40.0*i/27.0;
		else
			dB = -50.0 + (float) 50.0*(i-27)/100.0;

		volume_table[i] = pow(10.0, dB/20.0);
	}
	volume_table[0] = 0;

	panel = panel_init();
	if (!panel)
		error("ERROR: cannot access the panel.\n");

	status = rtems_timer_create(rtems_build_name('V', 'O', 'L', 'T'), &timer_id);
	if (status != RTEMS_SUCCESSFUL)
		error("Can't create the volume timer.\n");
	status = rtems_timer_fire_after(timer_id, 25, timer_routine, (void *) argument);
	if (status != RTEMS_SUCCESSFUL)
		error("Can't start the volume timer.\n");
		
	display(panel);

	while (TRUE)
	{
		if (!panel_in_new_event(panel, TRUE))
		{
			TRACE("ERROR: unexpected exit waiting new panel event.\n");
			break;
		}

		if (panel_in_potentiometer(panel, &potentiometer, &value))
		{
			if (potentiometer == PANEL01_POT_VOLUME)
			{
				parameters[kParam_Volume] = volume_table[value];
				rtems_timer_reset(timer_id);
			}
		}
		else if (panel_in_encoder(panel, &encoder, &increment))
		{
			change_param_value(param_id, increment);
			display(panel);
			rtems_timer_reset(timer_id);
		}
		else if (panel_in_keypad(panel, &key_bits))
		{
			if (key_bits & PANEL01_KEY_GROUP_UP)
			{
				if (menu.group_selected < menu.group_count-1)
				{
					menu.group_selected++;
					group = &menu.groups[menu.group_selected];
					page = &group->pages[group->page_selected];
					param_id = page->params[page->param_selected];
				}
			}
			else if (key_bits & PANEL01_KEY_GROUP_DOWN)
			{
				if (menu.group_selected > 0)
				{
					menu.group_selected--;
					group = &menu.groups[menu.group_selected];
					page = &group->pages[group->page_selected];
					param_id = page->params[page->param_selected];
				}
			}
			else if (key_bits & PANEL01_KEY_PAGE_UP)
			{
				if (group->page_selected < group->page_count-1)
				{
					group->page_selected++;
					page = &group->pages[group->page_selected];
					param_id = page->params[page->param_selected];
				}
			}
			else if (key_bits & PANEL01_KEY_PAGE_DOWN)
			{
				if (group->page_selected > 0)
				{
					group->page_selected--;
					page = &group->pages[group->page_selected];
					param_id = page->params[page->param_selected];
				}
			}
			else if (key_bits & PANEL01_KEY_PARAM_UP)
			{
				if (page->param_selected < page->param_count-1)
				{
					page->param_selected++;
					param_id = page->params[page->param_selected];
				}
			}
			else if (key_bits & PANEL01_KEY_PARAM_DOWN)
			{
				if (page->param_selected > 0)
				{
					page->param_selected--;
					param_id = page->params[page->param_selected];
				}
			}
			else if (key_bits & PANEL01_KEY_VALUE_UP)
			{
				change_param_value(param_id, 1);
				rtems_timer_reset(timer_id);
			}
			else if (key_bits & PANEL01_KEY_VALUE_DOWN)
			{
				change_param_value(param_id, -1);
				rtems_timer_reset(timer_id);
			}
						
			display(panel);
		}
	}

	panel_exit(panel);

	TRACE("Ending panel_task...\n");
	rtems_task_delete(RTEMS_SELF);
}

rtems_id create_panel_task(rtems_id dsp_task_id)
{
	rtems_id		task_id;
	rtems_status_code	status;

	status = rtems_task_create(
		rtems_build_name('T', 'H', 'R', 'U'),
		50,
		RTEMS_MINIMUM_STACK_SIZE,
		RTEMS_DEFAULT_MODES,
		RTEMS_DEFAULT_ATTRIBUTES,
		&task_id
	);

	if (status != RTEMS_SUCCESSFUL)
	{
		TRACE("ERROR: cannot create panel_task.\n");
		return 0;
	}

	status = rtems_task_start(task_id, panel_task, (rtems_task_argument) dsp_task_id);
	if (status != RTEMS_SUCCESSFUL)
	{
		TRACE("ERROR: cannot start panel_task.\n");
		return 0;
	}

	return task_id;
}

rtems_task rtems_main(rtems_task_argument ignored)
{
	rtems_id dsp_task_id;
	
	TRACE("******************************************************************************\n");
	TRACE("simpleverb: Reverberation with parameter control from ColdFire\n");
	TRACE("******************************************************************************\n");

	dsp_task_id = create_dsp_task();
	create_panel_task(dsp_task_id);
	rtems_task_delete(RTEMS_SELF);
}
